// This is part of the WAR SOFTWARE SERIES initiated by Jarle Aase
// Copyright 1996 by Jarle Aase. All rights reserved.
// See the "War Software Series Licende Agreement" for details concerning 
// use and distribution.
// ---
// This source code, executables and programs containing source code or
// binaries or proprietetary technology from the War Software Series are
// NOT alloed used, viewed or tested by any governmental agencies in
// any countries. This includes the government, departments, police, 
// military etc.
// ---
// This file is intended for use with Tab space = 2
// Created and maintained in MSVC Developer Studio
// ---
// NAME			:	VfFsys.cpp
// PURPOSE	:	War Daemon Virtual File System
// PROGRAM	: 
// DATE			:	March 13 1997
// AUTHOR		:	Jarle Aase
// ---
//
// Exports are done in the .def file.
// 
// REVISION HISTORY
// 

#include "stdafx.h"
#include <afxdllx.h>
#include <ctype.h>

#define EXT

// For now, we don't support non-daemon programs
//#include "WarSoftware.h"
#include "WarDaemon.h"
#include "WarFsys.h"
#include "VfFSys.h"
#include "UnixFsysTypes.h"
#include "FsysSecurity.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static AFX_EXTENSION_MODULE VfSysDLL = { NULL, NULL };


extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH)
	{
		TRACE0("VFSYS.DLL Initializing!\n");
		
		// Extension DLL one-time initialization
		AfxInitExtensionModule(VfSysDLL, hInstance);

		// Insert this DLL into the resource chain
		new CDynLinkLibrary(VfSysDLL);

		// Create therad sync object
		pLock = new CRWLock;
		if (!pLock->Create("VfSyslck"))
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: DllMain() - Unable to create CRWLock object.");
			return 0;
		}
	}
	else if (dwReason == DLL_PROCESS_DETACH)
	{
		TRACE0("VFSYS.DLL Terminating!\n");

		if (pVfSys)
			delete pVfSys;
		
		pVfSys = NULL;
		
		if (pLock)
			delete pLock;
		pLock = NULL;
		
		if (pSmem)
			delete pSmem;
		pSmem = NULL;
	}
	return 1;   // ok
}


////////////////////////////////////////////////////////////////////////////////
// Entry point

BOOL OpenFsys(HWND *phWnd, LPCSTR IniFilePath)
{
	int Index;

	if (pVfSys)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - File system is already initialized.");
		SetLastError(ERROR_ALREADY_EXISTS);
		return FALSE;
	}

	// Initialize the security subsystem
	// Create a new root security file it it don't already exist.
	{
		HANDLE h = CFsysSecurity::LoadSecurityGroup("",FALSE);

		if (h == INVALID_HANDLE_VALUE)
			h = CFsysSecurity::OpenSecurityGroupForUpdate("");

		if (h == INVALID_HANDLE_VALUE)
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - Failed initialize the security subsystem.");
			return FALSE;
		}
		
		CFsysSecurity::CloseSecurityGroup(h);
	}

	pVfSys = new CVirtualFileSystem;
	if (!pVfSys->CreateEx(0, AfxRegisterWndClass(0),
		_T("Wirtual File System Notification Sink"),
		WS_OVERLAPPED, 0, 0, 0, 0, NULL, NULL))
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - Unable to create Virtual File System notify window!");
		return FALSE;
	}
	ASSERT(pVfSys->m_hWnd != NULL);
	ASSERT(CWnd::FromHandlePermanent(pVfSys->m_hWnd) == pVfSys);

	*phWnd = pVfSys->m_hWnd;

	pLock->WriteLock();

	// Create shared memory
	pSmem = new CSharedBlksMem;
	if (!pSmem->Create("VfSysmem", 
		GetPrivateProfileInt("fsys", "buffer", (1024 * 512), IniFilePath)))
	{
		delete pVfSys;
		pVfSys = NULL;
		phWnd = NULL;
		CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - Unable to create shared memory");
		pLock->UnlockWrite();
		return FALSE;
	}

	// Create the master virtual file system
	if (pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_ROOTNODE] == 0)
	{
		// Create global headers
		SMEM GlobalHdr = *(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_DEVDRV]);
		if (GlobalHdr == 0)
		{
			if (!pSmem->sh_malloc(GlobalHdr, sizeof(GLOBAL_FSYS_DRIVERS_HDR)))
			{
				CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - Failed to allocate GLOBAL_FSYS_DRIVERS_HDR");
				goto fail;
			}
			*(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_DEVDRV]) = GlobalHdr;
		}

		// Create the global plugin table
		ASSERT(pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_PLUGINS] == 0);
		{
			SMEM smPluginTable;
			smPluginTable = 0;
			if (!pSmem->sh_malloc(smPluginTable, sizeof(GLOBAL_VFSYS_PLUGIN_HDR)))
			{
				CLog::GetLog()->LogMsg(LOGF_ERROR,"CVirtualFileSystem::OpenFsys() - Failed to allocate GLOBAL_VFSYS_PLUGIN_HDR");
				goto fail;
			}
			*(SMEM *)&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_PLUGINS] = smPluginTable;

			pGlobalPluginHdr = (GLOBAL_VFSYS_PLUGIN_HDR *)pSmem->__Ptr(smPluginTable);
		}
	
		for(Index = 0; Index < PLUGIN_TABLE_SIZE; Index++)
		{
			CString Tag, Name, Pattern;
			Tag.Format("Plugin%d", Index);
			
			GetPrivateProfileString(Tag,"Plugin","", Name.GetBuffer(32), 32, IniFilePath);
			Name.ReleaseBuffer();
			if (Name.IsEmpty())
				break;

			GetPrivateProfileString(Tag,"Pattern","", Pattern.GetBuffer(64), 64, IniFilePath);
			Pattern.ReleaseBuffer();
			
			strcpy(pGlobalPluginHdr->Plugin[Index].Name, Name);
			strcpy(pGlobalPluginHdr->Plugin[Index].Pattern, Pattern);
			pGlobalPluginHdr->Plugin[Index].dwFlags = GetPrivateProfileInt(Tag,"Flags",0, IniFilePath);
		}

		SyncPlugins(); // We only need to do this once
	
		if (!CreateMasterFsys(IniFilePath))
		{
fail:
			delete pVfSys;
			*phWnd = NULL;
			pVfSys = NULL;
			delete pSmem;
			pSmem = NULL;
			pLock->UnlockWrite();
			return FALSE;
		}
	}
	else
	{
		// Initialize only..
		SyncPlugins();
		SyncDeviceDrivers();
	}


	pLock->UnlockWrite();
	return TRUE;
}


// End of entry point

void SyncPlugins()
{
	pGlobalPluginHdr = (GLOBAL_VFSYS_PLUGIN_HDR *)pSmem->__Ptr(
		*(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_PLUGINS]));

	if (pGlobalPluginHdr)
	{
		for(int Index = 0; Index < PLUGIN_TABLE_SIZE; Index++)
		{
			if (!pGlobalPluginHdr->Plugin[Index].Name[0])
				break;
			if (!LocalPluginTable[Index].hDll)
			{
				// Load it
				// Try to load the device driver
				if ((LocalPluginTable[Index].hDll = LoadLibrary(pGlobalPluginHdr->Plugin[Index].Name)) == NULL)
				{
					CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: SyncPlugins() - Failed to load plugin'%s'. %s",
						pGlobalPluginHdr->Plugin[Index].Name, GetLastErrorText());
					continue;
				}

				// Find function(s)
				LocalPluginTable[Index].OnFoundFile = 
					(DWORD (*)(
									SMEM smDir, // Parent directory
									CFileInfo& Info, // File info from device
									SMEM smNode, // Node, 0 if new file
									CString& RealName, //File system name
									CString& VisualName, // The name the user will see
									CString& LinkTo, // Link to
									LPCSTR FsysPath, // File system path to parent dir
									FILE_SYSTEM *pFsys, // Current file system
									CDeviceDriver *pDev // Current device
									))GetProcAddress(
					LocalPluginTable[Index].hDll, 
					"?OnFoundFile@@YAKUSMEM@@AAVCFileInfo@@0AAVCString@@22PBDPAUFILE_SYSTEM@@PAVCDeviceDriver@@@Z");

				CLog::GetLog()->LogMsg(LOGF_DEBUG,"VfSys.dll: SyncPlugins() - loaded plugin '%s'",
					pGlobalPluginHdr->Plugin[Index].Name);
			}
		}
	}
}

// Mount the master file system
BOOL CreateMasterFsys(LPCSTR IniFilePath)
{
	int Index;

	for(Index = 0;; Index++)
	{
		CString Tag, Path, Vpath, Driver, User, Pwd;
		int Copy, Timeout;
		Tag.Format("mount%d", Index);

		GetPrivateProfileString(Tag,"Path","", Path.GetBuffer(1024), 1024, IniFilePath);
		Path.ReleaseBuffer();
		if (Path.IsEmpty())
			break;

		GetPrivateProfileString(Tag,"VPath","", Vpath.GetBuffer(1024), 1024, IniFilePath);
		Vpath.ReleaseBuffer();
		if (Vpath.IsEmpty())
			break;

		GetPrivateProfileString(Tag,"Driver","", Driver.GetBuffer(1024), 1024, IniFilePath);
		Driver.ReleaseBuffer();
		if (Driver.IsEmpty())
			break;

		GetPrivateProfileString(Tag,"User","", User.GetBuffer(32), 32, IniFilePath);
		User.ReleaseBuffer();

		GetPrivateProfileString(Tag,"passwd","", Pwd.GetBuffer(32), 32, IniFilePath);
		Pwd.ReleaseBuffer();

		Copy = GetPrivateProfileInt(Tag,"Copy", FALSE, IniFilePath);
		Timeout = GetPrivateProfileInt(Tag,"Timeout", FALSE, IniFilePath);

		MountFsys(Path, Vpath, Driver, User, Pwd, Copy, Timeout, IniFilePath, Tag);
	}

	return Index > 0;
}

// Mount a file system
BOOL MountFsys(LPCSTR Path, LPCSTR Vpath, LPCSTR Driver, LPCSTR User, LPCSTR Pwd, BOOL Copy, int Timeout, LPCSTR IniFilePath, LPCSTR Tag)
{
	SMEM *pMasterNode = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_ROOTNODE]);
	int DevIndex = -1;

	CLog::GetLog()->LogMsg(LOGF_SYSTEM,"vfsys.dll: MountFsys(%s, %s, %s) - Mounting file system.",
		Path, Vpath, Driver);

	// Check if the driver is loaded
	SyncDeviceDrivers();
	if ((DevIndex = MountDeviceDriver(Driver)) == -1)
		return FALSE;

	if (*pMasterNode)
		ScanPath(Vpath, TRUE, FALSE, TRUE);


	// Mount the file system
	if (!strcmp(Vpath,"/"))
	{
		// Root node.
		if (*pMasterNode)
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: MountFsys(%s, %s) - Root node already exist.",
				Path, Vpath);
			SetLastError(ERROR_ALREADY_EXISTS);
			return FALSE;
		}
	}

	// Create the file system structure
	int HdrLen = sizeof(FILE_SYSTEM);
	HdrLen += (strlen(Path) +1);
	HdrLen += (strlen(Vpath) +1);
	HdrLen += (strlen(User) +1);
	HdrLen += (strlen(Pwd) +1);
	
	SMEM smFsys;
	if (!pSmem->sh_malloc(smFsys, HdrLen))
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: MountFsys(%s, %s) - Out of shared memory.",
			Path, Vpath);
		SetLastError(ERROR_NOT_ENOUGH_MEMORY);
		return FALSE;
	}

	FILE_SYSTEM *pFsysLast = (FILE_SYSTEM *)pSmem->__Ptr(*(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_FSYS]));
	FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smFsys);

	// Initialize data members
	strcpy(pFsys->DOSPath, Path);
	strcpy(pFsys->Vpath, Vpath);
	strcpy(pFsys->LoginName, User ? User : "");
	strcpy(pFsys->LoginPwd, Pwd ? Pwd : "");

	pFsys->DevIdx = DevIndex;
	pFsys->RootNode = 0;

	pFsys->Next = pFsys->RootNode;

	pFsys->FlushDelay = Timeout;
	pFsys->DoCopy = Copy;


	if (!CreateRootNode(smFsys))
		return FALSE;

	// Initialize
	if (!pFsysLast)
	{
		// Link as first in list
		SMEM *pShdr = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_FSYS]);
		*pShdr = smFsys;
	}
	else
	{
		// Find the last fsys and link to it
		while(pFsysLast->Next)
			pFsysLast = (FILE_SYSTEM *)pSmem->__Ptr(pFsysLast->Next);
		if (pFsysLast)
			pFsysLast->Next = smFsys;
	}

	// Check for additional parameters..
	if (!GetPrivateProfileInt(Tag,"ScanOnDemand", TRUE, IniFilePath))
		ScanPath(Vpath, FALSE, TRUE, TRUE);

	return TRUE;
}

// Syncronize the local driver table against the global table
BOOL SyncDeviceDrivers()
{
	SMEM *pSmGlobalHdr = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_DEVDRV]);
	GLOBAL_FSYS_DRIVERS_HDR *pGlobalHdr = (GLOBAL_FSYS_DRIVERS_HDR *)pSmem->__Ptr(*pSmGlobalHdr);
	ASSERT(AfxIsValidAddress(pGlobalHdr, sizeof(GLOBAL_FSYS_DRIVERS_HDR)));

	for(int Index = 0;Index < DRIVER_TABLE_SIZE; Index++)
	{
		if (!*pGlobalHdr->Name[Index])
			break;
		if (!LocalDeviceDrivers[Index].pDevDr)
			MountDeviceDriver(pGlobalHdr->Name[Index]);
	}

	return TRUE;
}

// Mount a new device driver (if don't exist)
// Return the index to the driver
int MountDeviceDriver(LPCSTR Driver)
{
	SMEM *pSmGlobalHdr = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_DEVDRV]);
	GLOBAL_FSYS_DRIVERS_HDR *pGlobalHdr = (GLOBAL_FSYS_DRIVERS_HDR *)pSmem->__Ptr(*pSmGlobalHdr);
	
	ASSERT(AfxIsValidString(Driver));
	ASSERT(AfxIsValidAddress(pGlobalHdr, sizeof(GLOBAL_FSYS_DRIVERS_HDR)));

	CLog::GetLog()->LogMsg(LOGF_SYSTEM,"vfsys.dll: MountDeviceDriver(%s) - Mounting device driver.",
		Driver);

	if (strlen(Driver) >= DRIVER_NAME_LEN)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: MountDeviceDriver() - Name of device '%s' is more than %d characters.",
			Driver, DRIVER_NAME_LEN - 1);
		return -1;
	}

	int Index = FindDeviceDriver(Driver);
	
	if (Index < 0)
	{
		// Find a new empty slot
		for(Index = 0;Index < DRIVER_TABLE_SIZE; Index++)
		{
			if (!*pGlobalHdr->Name[Index])
				break;
		}
		strcpy(pGlobalHdr->Name[Index], Driver);
	}

	if (Index >= DRIVER_TABLE_SIZE)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: MountDeviceDriver() - Device table is full (%d entries). Unable to add '%s'.",
			DRIVER_TABLE_SIZE, Driver);
		return -1;
	}

	// Try to load the device driver
	if ((LocalDeviceDrivers[Index].hDll = LoadLibrary(Driver)) == NULL)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: MountDeviceDriver() - Failed to load device driver '%s'. %s",
			Driver, GetLastErrorText());
		return -1;
	}

	HANDLE (*DevOpenDevice)(CFileSystemTemplate *pInfo);
	if ((DevOpenDevice = (HANDLE(*)(CFileSystemTemplate *pInfo))GetProcAddress(
		LocalDeviceDrivers[Index].hDll, 
		"?DevOpenDevice@@YAPAXPAVCFileSystemTemplate@@@Z")) == NULL)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: MountDeviceDriver() - Could not find 'DevOpenDevice in dll '%s'",
			Driver);
		FreeLibrary(LocalDeviceDrivers[Index].hDll);
		LocalDeviceDrivers[Index].hDll = NULL;
		return -1;
	}

	LocalDeviceDrivers[Index].pDevDr = new CDeviceDriver;
	if (DevOpenDevice(LocalDeviceDrivers[Index].pDevDr) == INVALID_HANDLE_VALUE)
	{
		CLog::GetLog()->LogMsg(LOGF_ERROR,"VfSys.dll: MountDeviceDriver() - Could not initialize dll '%s'",
			Driver);
		FreeLibrary(LocalDeviceDrivers[Index].hDll);
		LocalDeviceDrivers[Index].hDll = NULL;
		delete LocalDeviceDrivers[Index].pDevDr;
		LocalDeviceDrivers[Index].pDevDr = NULL;
		return -1;
	}

	return Index; // Error
}

int FindDeviceDriver(LPCSTR Name)
{
	SMEM *pSmGlobalHdr = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_DEVDRV]);
	GLOBAL_FSYS_DRIVERS_HDR *pGlobalHdr = (GLOBAL_FSYS_DRIVERS_HDR *)pSmem->__Ptr(*pSmGlobalHdr);
	ASSERT(AfxIsValidAddress(pGlobalHdr, sizeof(GLOBAL_FSYS_DRIVERS_HDR)));
	
	for(int Index = 0;Index < DRIVER_TABLE_SIZE; Index++)
	{
		if (!stricmp(pGlobalHdr->Name[Index], Name))
			return Index;
	}

	return -1; // Not found
}

///////////////////////////////////////////////////////////////////////////////////////
// VfSys functions

// Create a root node for a file system
BOOL CreateRootNode(SMEM smFsys)
{
	SMEM smDirNode;
	FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smFsys);

	// Extract the name
	LPCSTR NodeName = strrchr(pFsys->Vpath,'/');
	if (NodeName)
		++NodeName;
	else
		return FALSE;

	if (!CreateDirNode(smDirNode, NodeName, NULL, NULL, NULL, NULL))
		return FALSE;

	pFsys->RootNode = smDirNode;

	DIR_NODE *pDirNode = (DIR_NODE *)pSmem->__Ptr(smDirNode);
	
	// Initialize
	pDirNode->m_FileSystem = smFsys;
	pDirNode->FileNode.f_IsVirtual = TRUE;
	pDirNode->FileNode.Perms = 0777 | NODE_DIR | NODE_DRIVE;

	// Is this the global root node?
	if (!*NodeName)
	{
		SMEM *pMasterNode = (SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_ROOTNODE]);
		if (!*pMasterNode)
		{
			*pMasterNode = smDirNode;
		}
		else
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: CreateRootNode(%s) - Global root node alredy exist.", pFsys->Vpath);
			pSmem->sh_free(smDirNode);
			pFsys->RootNode = 0;
			return FALSE;
		}
	}
	else
	{
		// Find the node to link it to
		SMEM smLinkNode;
		CPathArray PathArray;
		PathArray.ParsePathToArray(pFsys->Vpath);
		CString MyName = "";
		if (PathArray.GetUpperBound() > 0)
		{
			MyName = PathArray[PathArray.GetUpperBound()];
			PathArray.RemoveAt(PathArray.GetUpperBound());
		}
		else
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: CreateRootNode(%s) - Invalid map node name.", pFsys->Vpath);
			pSmem->sh_free(smDirNode);
			return FALSE;
		}
		if (!FindNodeFromPath(smLinkNode, PathArray))
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: CreateRootNode(%s) - Virtual map node don't exist.", pFsys->Vpath);
			pSmem->sh_free(smDirNode);
			return FALSE;
		}
		SMEM smTest = FindNode(smLinkNode, MyName);
		if (smTest)
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"vfsys.dll: CreateRootNode(%s) - Virtual map node already exist.", pFsys->Vpath);
			pSmem->sh_free(smDirNode);
			return FALSE;
		}

		pDirNode->FileNode.f_IsVirtual = TRUE;

		// Finally, link it
		LinkToFather(smDirNode, smLinkNode);
	}
	
	return TRUE;
}

BOOL FindNodeFromPath(SMEM &smDir, CPathArray& PathArray)
{
	smDir = *(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_ROOTNODE]);

	for(int Index = 1; Index <= PathArray.GetUpperBound(); Index++)
	{
		smDir = FindNode(smDir, PathArray[Index]);
		if (!smDir)
			return FALSE;
	}
	return TRUE;
}

BOOL FindNodeFromPath(SMEM& smDir, LPCSTR Path)
{
	CPathArray PathArray;
	PathArray.ParsePathToArray(Path);
	return FindNodeFromPath(smDir, PathArray);
}

void CPathArray::ParsePathToArray(LPCSTR Path)
{
	// Break up the path in names
	CString cBuf = Path;
	LPSTR p = cBuf.GetBuffer(1);

	RemoveAll();
	Add(""); // Root node

	for(p = strtok(p, "/"); p; p = strtok(NULL, "/"))
		Add(p);
}

BOOL CreateDirNode(
									 SMEM& smDirNode, 
									 LPCSTR Name, 
									 LPCSTR RealName, 
									 LPCSTR LinkTo, 
									 SMEM *psmFather, 
									 LPCSTR Comment)
{
	if (!AllocDirNode(smDirNode,Name,RealName,LinkTo,psmFather,Comment))
		return FALSE;

	if (psmFather)
		LinkToFather(smDirNode, *psmFather);

	return TRUE;
}

BOOL CreateFileNode(
										SMEM& smNode, 
										LPCSTR Name, 
										LPCSTR RealName, 
										LPCSTR LinkTo, 
										SMEM *psmFather, 
										LPCSTR Comment)
{
	if (!AllocFileNode(smNode,Name,RealName,LinkTo,psmFather,Comment))
		return FALSE;

	if (psmFather)
		LinkToFather(smNode, *psmFather);

	return TRUE;
}

BOOL AllocDirNode(
									SMEM& smDirNode, 
									LPCSTR Name, 
									LPCSTR RealName, 
									LPCSTR LinkTo, 
									SMEM *psmFather, 
									LPCSTR Comment)
{
	DIR_NODE *pDirNode;
	int HdrLen = sizeof(DIR_NODE) + GetDataBufLen(Name, RealName, Comment, LinkTo); 

	if (!pSmem->sh_malloc(smDirNode, HdrLen))
		return FALSE;

	pDirNode = (DIR_NODE *)pSmem->__Ptr(smDirNode);
	memset(pDirNode, 0, sizeof(DIR_NODE));

	InitializeFileNode(&pDirNode->FileNode, (LPSTR )pDirNode->Buf, Name, RealName, LinkTo, Comment);

	pDirNode->f_DoScan = TRUE; // Security permissions is not set
	pDirNode->FileNode.Perms = NODE_DIR;

	return TRUE;
}

BOOL AllocFileNode(
									 SMEM& smNode, 
									 LPCSTR Name, 
									 LPCSTR RealName, 
									 LPCSTR LinkTo, 
									 SMEM *psmFather, 
									 LPCSTR Comment)
{
	FILE_NODE *pNode;
	int HdrLen = sizeof(FILE_NODE) + GetDataBufLen(Name, RealName, Comment, LinkTo); 
	

	if (!pSmem->sh_malloc(smNode, HdrLen))
		return FALSE;

	pNode = (FILE_NODE *)pSmem->__Ptr(smNode);
	memset(pNode, 0, sizeof(FILE_NODE));

	InitializeFileNode(pNode, (LPSTR)&pNode->Buf, Name, RealName, LinkTo, Comment);

	return TRUE;
}

DWORD GetDataBufLen(
										 LPCSTR Name, 
										 LPCSTR RealName, 
										 LPCSTR Comment,
										 LPCSTR LinkTo)
{
	DWORD HdrLen = 0;
	HdrLen += Name ? (strlen(Name) + 1) : 0;
	HdrLen += RealName ? (strlen(RealName) +1) : 0;
	HdrLen += LinkTo ? (strlen(LinkTo) +1) : 0;
	HdrLen += Comment ? (strlen(Comment) +1) : 0;

	return HdrLen;
}

// Change the name/realname/comment of a node. Simply reallocate the
// Buf to use SMEM, and disregard the old information.
// If any of the variables are NULL, use the old value.
BOOL ChangeNodeInformation(
													 FILE_NODE *pNode,
													 LPCSTR Name, 
													 LPCSTR RealName,
													 LPCSTR LinkTo, 
													 LPCSTR Comment)
{
	SMEM smOldBuf;
	smOldBuf = 0;
	if (pNode->f_BufIsSMEM)
		smOldBuf = pNode->Buf;

	if (Name == NULL)
		Name = pNode->Name;
	
	if (RealName == NULL)
		RealName = pNode->RealName;

	if (LinkTo == NULL)
		LinkTo = pNode->LinkTo;
	
	if (Comment == NULL)
		Comment = pNode->Comment;

	// Calculate header len
	int SegLen = GetDataBufLen(Name, RealName, Comment, LinkTo);

	SMEM smNewBuf;
	if (!pSmem->sh_malloc(smNewBuf, SegLen))
		return FALSE;

	LPSTR lpszBuf = (LPSTR)pSmem->__Ptr(smNewBuf);
	InitializeFileNode(pNode,lpszBuf,Name,RealName,LinkTo,Comment);
	pNode->Buf = smNewBuf;
	pNode->f_BufIsSMEM = TRUE;

	if (smOldBuf)
		pSmem->sh_free(smOldBuf);

	return TRUE;
}



// Used when a file is renamed, or when the comment change.
// The node is replaced at the current location.
// The call will fail if the file is locked or open.
BOOL ReplaceDirNode(
										SMEM& smOldNode,
										SMEM& smNewNode, 
										LPCSTR Name, 
										LPCSTR RealName, 
										LPCSTR Comment)
{
	DIR_NODE *pOldNode = (DIR_NODE *)pSmem->__Ptr(smOldNode);
	if (pOldNode->FileNode.f_OpenCnt
		|| pOldNode->FileNode.f_Deleted
		|| pOldNode->FileNode.f_Locked)
	{
		SetLastError(ERROR_SHARING_VIOLATION);
		CLog::GetLog()->LogMsg(LOGF_DEBUG,"ReplaceDirNode(%s) - File is busy or deleted. Failed.", 
			pOldNode->FileNode.Name);
		return FALSE;
	}

	if (!AllocDirNode(
		smNewNode, 
		Name, 
		RealName, 
		pOldNode->FileNode.LinkTo,
		&pOldNode->FileNode.Father,
		Comment))
	{
		return FALSE;
	}

	// Copy data members
	DIR_NODE *pNewNode = (DIR_NODE *)pSmem->__Ptr(smNewNode);
	CopyDirNode(pNewNode, pOldNode);

	// Relink it.
	FILE_NODE *pSon = (FILE_NODE *)pSmem->__Ptr(pNewNode->Son);
	FILE_NODE *pNext = (FILE_NODE *)pSmem->__Ptr(pNewNode->FileNode.Next);
	FILE_NODE *pPrev = (FILE_NODE *)pSmem->__Ptr(pNewNode->FileNode.Prev);
	DIR_NODE *pFather = (DIR_NODE *)pSmem->__Ptr(pNewNode->FileNode.Father);
	
	if (pPrev)
	{
		pPrev->Next = smNewNode;
		ASSERT(pFather->Son != smOldNode);
	}
	else if (pFather)
	{
		ASSERT(pFather->Son == smOldNode);
		pFather->Son = smNewNode;
	}

	if (pNext)
		pNext->Prev = smNewNode;

	if (pSon)
		pSon->Father = smNewNode;

	FreeFsysNode(smOldNode, FALSE);

	return TRUE;
}

void CopyDirNode(DIR_NODE *pDstNode, DIR_NODE *pSrcNode)
{
	int i;

	CopyFileNode((FILE_NODE *)pDstNode, (FILE_NODE *)pSrcNode);

	for(i = 0; i < FSYS_DIR_MAX_UMASK; i++)
		pDstNode->m_DefaultPerms[i] = pSrcNode->m_DefaultPerms[i];

	pDstNode->m_FileSystem = pSrcNode->m_FileSystem;
	pDstNode->Timer = pSrcNode->Timer;
	pDstNode->Son = pSrcNode->Son;
	pDstNode->f_Dirty = pSrcNode->f_Dirty;
	pDstNode->f_DoScan = pSrcNode->f_DoScan;
}

void CopyFileNode(FILE_NODE *pDstNode, FILE_NODE *pSrcNode)
{
	pDstNode->Perms = pSrcNode->Perms;
	pDstNode->User = pSrcNode->User;
	pDstNode->Class = pSrcNode->Class;
	pDstNode->MoreUsers = pSrcNode->MoreUsers;
	pDstNode->Size = pSrcNode->Size;
	pDstNode->CreationTime = pSrcNode->CreationTime;
	pDstNode->LastWriteDate = pSrcNode->LastWriteDate;
	pDstNode->lastDlTime = pSrcNode->lastDlTime;
	pDstNode->DlCount = pSrcNode->DlCount;
	pDstNode->f_OpenCnt = pSrcNode->f_OpenCnt;
	pDstNode->f_Deleted = pSrcNode->f_Deleted;
	pDstNode->f_DeleteFile = pSrcNode->f_DeleteFile;
	pDstNode->f_Locked = pSrcNode->f_Locked;
	pDstNode->f_UseCopy = pSrcNode->f_UseCopy;
	pDstNode->f_UseHome = pSrcNode->f_UseHome;
	pDstNode->f_CopyToHome = pSrcNode->f_CopyToHome;
	pDstNode->f_PseudoName = pSrcNode->f_PseudoName;
	pDstNode->f_IsVirtual = pSrcNode->f_IsVirtual;
	pDstNode->Next = pSrcNode->Next;
	pDstNode->Prev = pSrcNode->Prev;
	pDstNode->Father = pSrcNode->Father;
}

// Always link as first
void LinkToFather(SMEM smNode, SMEM smFather)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smFather);
	ASSERT(pDir->FileNode.Perms & NODE_DIR); // Father must be directory

	FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(smNode);
	FILE_NODE *pNext = (FILE_NODE *)pSmem->__Ptr(pDir->Son);

	pNode->Next = pDir->Son;
	if (pNext)
		pNext->Prev = smNode;

	pDir->Son = smNode;
	pDir->f_DoScan = TRUE;
	pNode->Father = smFather;
}

// Initialize the strings in a file node
void InitializeFileNode(
									 FILE_NODE *pNode, 
									 LPSTR Buf, 
									 LPCSTR Name, 
									 LPCSTR RealName, 
									 LPCSTR LinkTo,
									 LPCSTR Comment)
{
	if (Name && RealName && !stricmp(Name,RealName))
		RealName = NULL;

	if (Name && *Name){pNode->Name = Buf;while(*Name)	*Buf++ = *Name++;	*Buf++ = 0;} else pNode->Name = NULL;
	if (RealName && *RealName){pNode->RealName = Buf;while(*RealName)	*Buf++ = *RealName++;	*Buf++ = 0;} else pNode->RealName = NULL;
	if (LinkTo && *LinkTo){pNode->LinkTo = Buf;while(*LinkTo)	*Buf++ = *LinkTo++;	*Buf++ = 0;} else pNode->LinkTo = NULL;
	if (Comment && *Comment){pNode->Comment = Buf;while(*Comment)	*Buf++ = *Comment++;	*Buf++ = 0;} else pNode->Comment = NULL;
	if (!pNode->RealName)
		pNode->RealName = pNode->Name;
}

// Prepere a path for access. (Primary VfSys refresh/security info routine)
// If FullScanToPath is true, scan/refresh the dirs down to the destrination
// If Recursive is true, scan all dirs from destination and down
BOOL ScanPath(LPCSTR Path, BOOL FullScanToPath, BOOL Recursive, BOOL Force)
{
	SMEM smDir = *(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_ROOTNODE]);
	
	// /
	// /C
	// /C/fsys/test/hello

	CPathArray Paths;
	Paths.ParsePathToArray(Path);

	for(int Index = 0;; Index++)
	{
		CString Name = Paths[Index];
		LPCSTR Pattern;
		BOOL ProcessLink = FALSE;
		if (FullScanToPath || (Index == Paths.GetUpperBound()))
			Pattern = NULL;
		else
			Pattern = Paths[Index +1];

		// Scan the directory
		ScanDir(smDir, Pattern, Force);

		// Find next dir to scan
		DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);

		if (pDir && (pDir->FileNode.Perms & NODE_DIR))
		{
			if (Index < Paths.GetUpperBound())
			{
				smDir = FindNode(smDir, Paths[Index +1]);
				pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
				if (!pDir)
					break; // Not a directory

				if (pDir && (pDir->FileNode.Perms & NODE_LINK))
				{
					// Follow link
					if (pDir->FileNode.LinkTo && pDir->FileNode.LinkTo[0])
					{
						if (pDir->FileNode.f_LinkNotResolved)
						{
							// The link is a DOS path. Resolve it to VfSys.
							CString VfSysPath = ResolveVfSysPathFromDosPath(pDir->FileNode.LinkTo);
							if (VfSysPath.IsEmpty())
							{
								CLog::GetLog()->LogMsg(LOGF_WARNINGS,"VfSys.dll: ScanPath() - Unresolved link: '%s'", 
									pDir->FileNode.LinkTo);
								return FALSE;
							}
					
							ChangeNodeInformation((FILE_NODE *)pDir,NULL,NULL,VfSysPath,NULL);
							pDir->FileNode.f_LinkNotResolved = FALSE;
							ProcessLink = TRUE;
						}

						CString LinkPath = pDir->FileNode.LinkTo;
						if (LinkPath[LinkPath.GetLength() -1] == '/')
							LinkPath.ReleaseBuffer(LinkPath.GetLength() -1);
						++Index;
						while(++Index <= Paths.GetUpperBound())
						{
							LinkPath += '/';
							LinkPath += Paths[Index];
						}
						BOOL Rval =  ScanPath(LinkPath, FullScanToPath, Recursive, Force);
						if (ProcessLink)
						{
							// Find out if this is a file or a directory
							SMEM smLinkToNode;
							if (FindNodeFromPath(smLinkToNode, pDir->FileNode.LinkTo))
							{
								FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(smLinkToNode);
								if (pNode && (pNode->Perms & NODE_DIR))
									pDir->FileNode.Perms |= NODE_DIR;
								else
									pDir->FileNode.Perms &= ~NODE_DIR;
							}
						}
						return Rval;
					}
					else
					{
						CLog::GetLog()->LogMsg(LOGF_WARNINGS,"VfSys.dll: ScanPath() - Missing link name in path %s (node=%d)", Path, Index);
						return FALSE;
					}
				}
				else if ((pDir->FileNode.Perms & NODE_DIR) == 0)
					break;
			}
			else
				break;
		}
	}

	if (Recursive)
		ScanDown(smDir, Force);

	return TRUE;
}

// Scan all nodes below the current node
void ScanDown(SMEM smDir, BOOL Force)
{
	while(smDir)
	{
		DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
		if (pDir->FileNode.Perms & NODE_DIR)
		{
			if (!pDir->FileNode.f_Deleted && !(pDir->FileNode.Perms & NODE_LINK))
			{
				LPCSTR p = pDir->FileNode.Name;
				if (!(p && (p[0] == '.') && (!strcmp(p,".") || !strcmp(p,".."))))
				{
					ScanDir(smDir,"*", Force);
					if (pDir->Son)
						ScanDown(pDir->Son, Force);
				}
			}
		}
		smDir = pDir->FileNode.Next;
	}
}

// Look for the name in the directory
SMEM FindNode(SMEM& smDir, LPCSTR Name)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
	SMEM Rval;

	Rval = 0;

	if (!pDir)
		return Rval;

	FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(pDir->Son);
	if (pNode)
	{
		Rval = pDir->Son;
		while(pNode)
		{
			if (!pNode->f_Deleted)
			{
				if (pNode->Name && !stricmp(pNode->Name, Name))
					break;
				if (pNode->RealName && !stricmp(pNode->RealName, Name))
					break;
			}
			Rval = pNode->Next;
			pNode = (FILE_NODE *)pSmem->__Ptr(Rval);
		}
	}
	
	return Rval;
}

BOOL ScanDir(SMEM smDir, LPCSTR Pattern, BOOL Force)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
	CDeviceDriver *pDev = NULL;
	BOOL HaveSecHandle = FALSE;
	HANDLE h;
	CString UseName;
	BOOL TryExact;
	
	if (!(pDir->FileNode.Perms & NODE_DIR))
		return FALSE; // Not a directory

	SMEM smFsys;
	CString FsysPath = ResolveFsysPath(smDir, smFsys);
	FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smFsys);
	if (!pFsys)
		return FALSE;

	{
		// See if we have to scan...
		if (!Force && !pDir->f_DoScan &&
			(!pFsys->FlushDelay || !pDir->Timer.TimeOut(pFsys->FlushDelay * 1000)))
			return TRUE; // No need to scan..

		if (!Pattern || !*Pattern)
				Pattern = "*";

		TryExact = ((Pattern[0] == '*') && !Pattern[1]) == 0;

		MarkAllNodesAsNotRefreshed(pDir->Son, Pattern);

		pDev = LocalDeviceDrivers[pFsys->DevIdx].pDevDr;
		if (pDev)
		{
			CFileInfo Info;

			if (!FsysPath.IsEmpty() && (FsysPath[FsysPath.GetLength() -1] != '/'))
				FsysPath += '/';
			FsysPath += Pattern;

#ifdef _DEBUG
		CLog::GetLog()->LogMsg(LOGF_DEBUG,"VfSys.dll: ScanDir() - rescanning '%s'", FsysPath);
#endif

			if ((h = pDev->FindFirstFile(FsysPath, Info)) == INVALID_HANDLE_VALUE)
			{
				if (TryExact)
				{
					// Give the plugins a chance. Try each known extention
					CString OriginalFsysPath = FsysPath;
					for(int Index = 0; LocalPluginTable[Index].hDll != NULL; Index++)
					{
						if (pGlobalPluginHdr->Plugin[Index].Pattern[0])
						{
							LPSTR p = strrchr(pGlobalPluginHdr->Plugin[Index].Pattern,'*');
							if (p && *++p)
							{
								FsysPath = OriginalFsysPath;
								FsysPath += p;
								if ((h = pDev->FindFirstFile(FsysPath, Info)) != INVALID_HANDLE_VALUE)
									goto match;
							}
						}
					}
				}
			}
			else // Match
			{
match:
				HaveSecHandle = TRUE;
				do
				{
					DWORD PluginFlags = 0;
					SMEM smNode = FindNode(smDir, Info.m_Name);
					DIR_NODE *pNewNode;
					CString RealName = Info.m_Name, VisualName = Info.m_Name, LinkTo = "";

					// Plugin
					for(int Index = 0; LocalPluginTable[Index].hDll != NULL; Index++)
					{
						if (LocalPluginTable[Index].OnFoundFile
							&& PatternMatchesName(RealName, pGlobalPluginHdr->Plugin[Index].Pattern))
						{
							PluginFlags |= LocalPluginTable[Index].OnFoundFile(
								smDir, Info, smNode, RealName, VisualName, LinkTo, FsysPath, pFsys, pDev);
						}
					}

					// Go on with processing

					if (smNode)
					{
						// Refresh
						pNewNode = (DIR_NODE *)pSmem->__Ptr(smNode);
						if (!pNewNode->FileNode.f_OpenCnt 
							&& !pNewNode->FileNode.f_Locked
							&& CompareFileTime(&pNewNode->FileNode.CreationTime, Info.GetCreationTime()))
						{
							// Not the same node. Unlink the old one and cerate a new one...
							DeleteNode(smNode);
							goto add_new_node;
						}
						PluginFlags &= ~LINK_PRESENT_NOT_RESOLVED; // Only needed when added/re-added
					}
					else
					{
add_new_node:
						// Add
						if (RealName == VisualName)
							RealName.Empty();
						if (Info.IsDirectory())
						{
							if (!CreateDirNode(smNode, VisualName, RealName, LinkTo, &smDir, Info.GetComment()))
							{
								pDev->FindClose(h);
								return FALSE;
							}
						}
						else
						{
							if (!CreateFileNode(smNode, VisualName, RealName, LinkTo, &smDir, Info.GetComment()))
							{
								pDev->FindClose(h);
								return FALSE;
							}
						}
						pNewNode = (DIR_NODE *)pSmem->__Ptr(smNode);
					}
					pNewNode->FileNode.Perms = Info.GetPermissions();
					pNewNode->FileNode.f_IsRefreshed = TRUE;
					pNewNode->FileNode.f_UseCopy = pFsys->DoCopy;
					pNewNode->FileNode.CreationTime = *Info.GetCreationTime();
					pNewNode->FileNode.LastWriteDate = *Info.GetModifyTime();
					pNewNode->FileNode.Size = Info.GetFileLength();
					pNewNode->FileNode.User = Info.GetUser();
					pNewNode->FileNode.Class = Info.GetGroup();
					pNewNode->FileNode.DlCount = Info.GetDLCount();

					// Check plugin flags
					if (PluginFlags & LINK_PRESENT_NOT_RESOLVED)
						pNewNode->FileNode.f_LinkNotResolved = TRUE;

					if (Info.IsDirectory())
					{
						pNewNode->SpecialNewFilePerm = Info.GetSpecialNewFilePerms();
					}

					if (pNewNode->FileNode.Perms & NODE_LINK 
						&& pNewNode->FileNode.Name 
						&& pNewNode->FileNode.Name[0])
						pNewNode->FileNode.Perms |= 0777; // Give all perms to soft link - they are ignored

				} while(pDev->FindNextFile(h, Info));
			}
		}
	}

	// Delete files that don't exist
	FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(pDir->Son);
	if (pNode)
	{
		SMEM smCurr = pDir->Son;
		while(pNode)
		{
			SMEM smNext = pNode->Next;
			if (!pNode->f_IsRefreshed && !pNode->f_IsVirtual && pNode->Name)
			{
				DeleteNode(smCurr);
			}
			else if (HaveSecHandle && pNode->f_IsVirtual && pNode->Name)
			{
				HANDLE hGrp = pDev->GetSecurityHandleFromFindHandle(h);
				if (hGrp)
				{
					CFileInfo Info;
					Info.m_Name = pNode->Name;
					Info.m_Flags = 0;
					if (CFsysSecurity::GetFsysSecurityProperties(hGrp, Info))
					{
						ASSERT(pNode->Perms & NODE_DIR);
						pNode->Perms &= ~NODE_SETTABLE;
						pNode->Perms |= Info.GetPermissions();
						pNode->f_IsRefreshed = TRUE;
						pNode->User = Info.GetUser();
						pNode->Class = Info.GetGroup();
						pNode->DlCount = Info.GetDLCount();
						if (pNode->Perms & NODE_DIR)
							((DIR_NODE *)pNode)->SpecialNewFilePerm = Info.GetSpecialNewFilePerms();
					}
				}
			}
			smCurr = smNext;
			pNode = (FILE_NODE *)pSmem->__Ptr(smNext);
		}
	}

	if (HaveSecHandle)
		pDev->FindClose(h);

	if (*Pattern == '*')
	{
		pDir->f_DoScan = FALSE;
		pDir->Timer.Reset();
	}

	return TRUE;
}

void MarkAllNodesAsNotRefreshed(SMEM smNode, LPCSTR Pattern)
{
	while(smNode)
	{
		FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(smNode);
		
		if (!pNode->f_IsVirtual && pNode->RealName)
		{
			if (PatternMatchesName(pNode->RealName, Pattern) 
				|| (pNode->Name && (pNode->Name != pNode->RealName) && PatternMatchesName(pNode->Name, Pattern)))
				pNode->f_IsRefreshed = FALSE;
		}
		
		smNode = pNode->Next;
	}
}

// Get the full path so that we can call the device driver 
CString ResolveFsysPath(SMEM smNode, SMEM& smFsys)
{
	CString ReversePath = "";
	LPCSTR p;
	BOOL GotFileSystem = FALSE;

	smFsys = 0;

	while(smNode)
	{
		DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smNode);
		LPCSTR Name = p = pDir->FileNode.RealName ? pDir->FileNode.RealName : pDir->FileNode.Name;
		
		if (Name && *Name && !ReversePath.IsEmpty())
			ReversePath += '/';

		if ((pDir->FileNode.Perms & NODE_DIR))
		{
			if (pDir->m_FileSystem)
			{
				smFsys = pDir->m_FileSystem;
				GotFileSystem = TRUE;
				break;
			}
		}

		if (Name && *Name)
		{
			p += strlen(Name) - 1;
			while(p >= Name)
			{
				ReversePath += *p;
				--p;
			}
		}

		smNode = pDir->FileNode.Father;
	}
	
	FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smFsys);
	if (pFsys)
	{
		LPCSTR Name = p =  pFsys->DOSPath;
		if (*Name)
		{
			p += strlen(Name) -1;
			if ((*p != '\\') && (*p != '/'))
				ReversePath += '/';
			while(p >= Name)
			{
				if (*p == '\\')
					ReversePath += '/';
				else
					ReversePath += *p;

				*p--;
			}
		}
	}

	// Reverse the string
	ReversePath.MakeReverse();
	return ReversePath;
}

CString ResolveVfSysPathFromDosPath(LPCSTR DOSPath)
{
	SMEM smBestFsys, smTestFsys;
	int nBestFsys = 0;
	smBestFsys = 0;
	CString MyPath = "";
	
	// Get the best match
	smTestFsys = *(SMEM *)(&pSmem->m_Segments[0].m_pSMH->pUserPtr[USERPTR_FSYS]);
	while(smTestFsys)
	{
		int cnt = 0;
		FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smTestFsys);
		if (!pFsys)
		{
			CLog::GetLog()->LogMsg(LOGF_ERROR,"fSys.dll: ResolveVfSysPathFromDosPath() - Unresolved fsys (SMEM = %d)", 
				(int)smTestFsys);
			return "";
		}
		LPCSTR p = DOSPath;
		LPSTR pp = pFsys->DOSPath;

		while(*p && tolower(*p) == tolower(*pp))
		{
			++p; ++pp; ++cnt;
		}
		if ((cnt > nBestFsys) && !*pp && ((*p == '\\') || (*p == '/') || (*p == 0)))
		{
			nBestFsys = cnt;
			smBestFsys = smTestFsys;
		}

		smTestFsys = pFsys->Next;
	}

	if (!smBestFsys)
		return "";

	FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smBestFsys);
	ASSERT(pFsys != NULL);
	MyPath = pFsys->Vpath;
	if (MyPath.GetLength() 
		&& (MyPath[MyPath.GetLength() -1] != '/') 
		&& (MyPath[MyPath.GetLength() -1] != '\\'))
		MyPath += '/';

	if ((*(DOSPath + nBestFsys) == '/')
		|| (*(DOSPath + nBestFsys) == '\\'))
		++nBestFsys;
	
	MyPath += DOSPath + nBestFsys;
	LPSTR p = MyPath.GetBuffer(1);
	while(*p)
	{
		if (*p == '\\')
			*p++ = '/';
		else
			++p;
	}
	MyPath.ReleaseBuffer();
	return MyPath;
}

// Delete a node (or mark it as deleted if it is in use)
void DeleteNode(SMEM smNode)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smNode);
	if (pDir->FileNode.Perms & NODE_DIR)
		DeleteTree(pDir->Son);
	pDir->FileNode.f_Deleted = TRUE;

	if (pDir->FileNode.f_OpenCnt == 0)
	{
		// Perform the delete. The node is 'free'
		UnlinkNode(smNode);

		if (pDir->FileNode.f_DeleteFile)
		{
			SMEM smFsys;

			CString cPath = ResolveFsysPath(smNode, smFsys);
			FILE_SYSTEM *pFsys = (FILE_SYSTEM *)pSmem->__Ptr(smFsys);
			if (pFsys)
			{
				CDeviceDriver *pDev = LocalDeviceDrivers[pFsys->DevIdx].pDevDr;
				if (pDev)
				{
					if (pDir->FileNode.Perms & NODE_DIR)
						pDev->RemoveDirectory(cPath);
					else
						pDev->DeleteFile(cPath);
				}
			}
		}

		// Free the memory
		FreeFsysNode(smNode);
	}
}

void UnlinkNode(SMEM smNode)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smNode);

	ASSERT(pDir->FileNode.f_OpenCnt == 0);

	FILE_NODE *pNext = (FILE_NODE *)pSmem->__Ptr(pDir->FileNode.Next);
	FILE_NODE *pPrev = (FILE_NODE *)pSmem->__Ptr(pDir->FileNode.Prev);
	DIR_NODE *pFather = (DIR_NODE *)pSmem->__Ptr(pDir->FileNode.Father);
		
	if (pNext)
		pNext->Prev = pDir->FileNode.Prev;
	if (pPrev)
		pPrev->Next = pDir->FileNode.Next;
	else if (pFather)
		pFather->Son = pDir->FileNode.Next;
}

// Free the memory used by a node.
// It is unlinked and 'dead'.
void FreeFsysNode(SMEM& smNode, BOOL KillExtraToo)
{
	FILE_NODE *pNode = (FILE_NODE *)pSmem->__Ptr(smNode);

	if (KillExtraToo)
	{
		if (pNode->MoreUsers)
			pSmem->sh_free(pNode->MoreUsers);

		if (pNode->f_BufIsSMEM)
			pSmem->sh_free(pNode->Buf);
	}

	pSmem->sh_free(smNode);
}

void DeleteTree(SMEM smDir)
{
	while(smDir)
	{
		DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
		if (pDir->FileNode.Perms & NODE_DIR)
			DeleteTree(pDir->Son);
		SMEM smCurr = smDir;
		smDir = pDir->FileNode.Next;
		DeleteNode(smCurr);
	}	
}


// Look for busy files at and below the object
// No need to expand (scan) the paths, as busy
// nodes will always reside in memory
BOOL ScanForBusyFiles(SMEM smDir)
{
	DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);
	
	while(smDir)
	{
		DIR_NODE *pDir = (DIR_NODE *)pSmem->__Ptr(smDir);

		if (pDir->FileNode.f_OpenCnt || pDir->FileNode.f_Locked)
			return TRUE;

		if (pDir->FileNode.Perms & NODE_DIR)
		{
			if (ScanForBusyFiles(pDir->Son))
				return TRUE;
		}
		smDir = pDir->FileNode.Next;
	}	

	return FALSE;
}

/////////////////////////////////////////////////////////////////
// CFileList

CFileList::~CFileList()
{
	while(GetAndDeleteFirst())
		;
}

BOOL CFileList::ListRaw(CString& cReturnBuf, DWORD Flags)
{
	cReturnBuf.Empty();

	FILE_NODE *pNode;
	CString cBuf;


	while(pNode = (FILE_NODE *)GetAndDeleteFirst())
	{
		cBuf.Empty();
		char buf[64];

		cBuf += '\"';
		cBuf += pNode->Name;
		cBuf += "\" ";

		*buf = 0;
		sprintf(buf, "%I64d ", pNode->Size);
		cBuf += buf;

		*buf = 0;
		sprintf(buf, "%I64x ", FiletimeToDate(pNode->LastWriteDate));
		cBuf += buf;
		
		*buf = 0;
		sprintf(buf, "%X ", pNode->Perms);
		cBuf += buf;

		*buf = 0;
		sprintf(buf, "%X ", pNode->User);
		cBuf += buf;

		*buf = 0;
		sprintf(buf, "%X ", pNode->Class);
		cBuf += buf;

		cBuf += '\"';
		if (pNode->Comment)
			cBuf += pNode->Comment;
		cBuf += "\"\r\n";

		cReturnBuf += cBuf;
	}

	return TRUE;
}

